module Wiki (
	Backend (getCurrent,getList,get,setCurrent,update,listKeys)
	,PGB
	,createPGB
	,Markup (Text, Paragraph, Font, Link, Heading, Url, Pre,List)
	,FontOp (Bold, Emph, Mono, Underline, Strike)
	,Document
	,wikiParser
	,toHtml
	,toLatex

) where

import Dbconnect
import Data.Char
import Text.ParserCombinators.Parsec

class Backend a where
	--Keyword -> (Full text,date)
	getCurrent :: a -> String -> IO (Maybe (String,String))

	--Keyword -> [(id,date,author,comment)]
	getList :: a -> String -> IO [(String, String, String, String)]

	--Keyword -> id -> Full text
	get :: a -> String -> String -> IO (Maybe String)

	--Keyword -> id -> ()
	setCurrent :: a -> String -> String -> IO Bool

	--Keyword -> Full text -> id
	update :: a -> String -> String -> String -> String -> IO String

	--[Keyword]
	listKeys :: a -> IO [String]

data Markup = Text String
	| Paragraph 
	| Link String String
	| Font FontOp [Markup]
	| Heading Int [Markup]
	| Url String
	| Pre [Markup]
	| List Bool [[Markup]]

data FontOp = Bold | Emph | Mono | Underline | Strike

type Document = [Markup]

wikiParser :: Parser Document	
wikiParser = many1 pMain

pPara :: Parser Markup	
pPara = count 2 pEol >> return (Paragraph)

pLink :: Parser Markup	
pLink = do
	string "[["
	(l,d) <- pLinkContent
	case d of
		[] -> return (Link l l)
		_ -> return (Link l d)

pList :: Bool -> Parser Markup
pList enum = do
	list <- many1 $ try $ pListItem enum
	return (List enum list)

pListItem :: Bool -> Parser [Markup]
pListItem enum = do
	many1 pEol
	char $ listToken enum
	many pOneLine

listToken True = '#'
listToken False = '*'
	
pHeading :: Parser Markup	
pHeading = do
	many1 pEol
	level <- many1 $ char '='
	s <- pStopAt level
	return (Heading (length level) s)

pURL :: Parser Markup	
pURL = do
	proto <- many1 letter
	string "://"
	s <- many1 (alphaNum <|> oneOf "?.:&-/~%=\\_")
	return (Url (proto++"://"++s))

pMail :: Parser Markup
pMail = do
	user <- many1 (alphaNum <|> oneOf ".-_")
	char '@'
	server <- many1 (alphaNum <|> oneOf ".-_")
	return (Url ("mailto:"++user++"@"++server))
	

pPre :: Parser Markup	
pPre = string "<pre>" >> do
	s <- pUntil (pOneLine <|> pEol) "</pre>"
	return (Pre s)

pBold,pEmph,pUnderline,pStrike :: Parser Markup	
pBold = pFont "**" Bold
pEmph = pFont "//" Emph
pUnderline = pFont "__" Underline
pStrike = pFont "--" Strike
pMono = pFont "||" Mono

pFont s o = string s >> do
	s <- pStopAt s
	return (Font o s)

pOtherChar :: Parser Markup	
pOtherChar = do
	c <- noneOf "\n\r"
	return (Text (c:[]))

pText :: Parser Markup	
pText = do
	t <- many1 alphaNum
	return (Text t)

pLinkContent :: Parser (String, String)	
pLinkContent = do try (string "]]" >> return ([],[])) 
	<|> try (do
		string " | "
		d <- pString "]]"
		return ([],d))
	<|> (do
		c <- anyChar
		(l,d) <- pLinkContent
		return (c:l,d))

pStopAt :: String -> Parser Document	
pStopAt xs = pUntil pOneLine xs

pString :: String -> Parser String	
pString xs = pUntil anyChar xs

pEol :: Parser Markup	
pEol = do
	(string "\n" <|> string "\r\n")
	return (Text "\r\n")

pUntil :: Parser a -> String -> Parser [a]	
pUntil p xs = manyTill p (try $ string xs)
	
pMain :: Parser Markup	
pMain = choice [
	try (pHeading)
	,try $ pList True
	,try $ pList False
	,try (pPara) 
	,try(pPre) 
	,try(pBold) 
	,try(pEmph) 
	,try(pUnderline) 
	,try(pStrike) 
	,try(pMono) 
	,try (pLink)
	,try (pURL)
	,try (pMail)
	,pText
	,pOtherChar
	,pEol
	]

pOneLine = choice [
	try pBold
	,try pEmph
	,try pUnderline
	,try pStrike
	,try pMono
	,try pLink
	,try pURL
	,try pMail
	,pText
	,pOtherChar
	]

newtype PGB = PGB DBService

createPGB :: String -> String -> String -> String -> IO PGB
createPGB host database user password = let db = createDBService host database "" user password Nothing in return (PGB db)

instance Backend PGB where

	getCurrent (PGB db) key = do 
		result <- selectReturnTuples db $ "SELECT fulltext,timestamp FROM curtexts WHERE keyword="++tov key
		case result of
			[text,date]:_ -> return (Just (text,date))
			_ -> return Nothing

	getList (PGB db) key = do
		list <- selectReturnTuples db $ "SELECT id, timestamp, author, comment from fulltexts WHERE keyword = "++tov key
		case list of
			[[]] -> return []
			_ -> return $ map (\[id,date,author,comment] -> (id,date,author,comment)) list

	get (PGB db) key id = do
		list <- selectReturnTuples db $ "SELECT fulltext from fulltexts WHERE id = "++tov id	
		case list of
			[s]:_ -> return (Just s)
			_ -> return Nothing

	setCurrent (PGB db) key id = do
		full <- selectReturnTuples db $ "SELECT keyword FROM fulltexts WHERE keyword="++tov key++" AND id='"++id++"'"
		cur <- selectReturnTuples db $ "SELECT keyword FROM curtexts WHERE keyword="++tov key
		case full of
			[[]] -> do
				return False
			_  -> do
				rows <- case cur of
					[[]] -> do
						execute db $ "INSERT INTO current (keyword, id) VALUES ("++tov key++","++tov id++")"
					_  -> do
						execute db $ "UPDATE current SET id = "++tov id++" WHERE keyword = "++tov key
				if rows == 1 then return True
					else return False 

	update (PGB db) key text author comment = do
		rows <- execute db $ "INSERT INTO fulltexts (keyword,fulltext, author, comment) VALUES ("++tov key++","++tov text++","++tov author++","++tov comment++")"
		if rows == 0 then return ""
			else do
				[[id]] <- selectReturnTuples db $ "SELECT currval('fulltexts_id_seq')" 
				setCurrent (PGB db) key id
				return id

	listKeys (PGB db) = do
		list <- selectReturnTuples db $ "SELECT keyword FROM current keyword ORDER BY lower(keyword)"
		case list of
			[[]] -> return []
			_ -> mapM (\[key] -> return key) list

tov :: String -> String
tov s = '\'':escapeQuery s++"'"
		
toHtml :: [Markup] -> String
toHtml [] = []
toHtml ((Paragraph):xs) = "<p>\n"++toHtml xs
toHtml ((Text s):xs) = s++toHtml xs
toHtml ((Link l d):xs) = "<link: "++l++" desc: "++d++">"++toHtml xs
toHtml ((Font o d):xs) = "<"++htmlFontOp o++">"++toHtml d++"</"++htmlFontOp o++">"++toHtml xs
toHtml ((Heading n d):xs) = "\n<h"++show n++">"++toHtml d++"</h"++show n++">\n"++toHtml xs
toHtml ((Url l):xs) = "<link: "++l++">"++toHtml xs
toHtml ((Pre s):xs) = "<pre>"++toHtml s++"</pre>"++toHtml xs
toHtml ((List o l):xs) = "<"++htmlListType o++">\n"++(unlines $ map (\s -> "<li>"++toHtml s++"</li>\n") l) ++ "</"++htmlListType o++">"++toHtml xs

htmlFontOp Bold = "b"
htmlFontOp Emph = "em"
htmlFontOp Mono = "tt"
htmlFontOp Underline = "u"
htmlFontOp Strike = "strike"

htmlListType True = "ol"
htmlListType False = "ul"

toLatex :: [Markup] -> String
toLatex [] = []
toLatex ((Paragraph):xs) = "\n\n"++toLatex xs
toLatex ((Text s):xs) = s++toLatex xs
toLatex ((Link l d):xs) = "{\\em "++d++"}"++toLatex xs
toLatex ((Font o d):xs) = "{\\"++latexFontOp o++" "++toLatex d++"}"++toHtml xs
toLatex ((Heading n d):xs) = "\n\\"++(unwords $ take (n-1) $ repeat "sub")++"section{"++toLatex d++"}"++"\n\n"++toLatex xs
toLatex ((Url l):xs) = "{\\bf "++l++"}"++toLatex xs
toLatex ((Pre s):xs) = "<pre>"++toLatex s++"</pre>"++toLatex xs
toLatex ((List o l):xs) = "\n\\begin{"++latexListType o++"}\n"++(unlines $ map (\s -> "\\item "++toLatex s++"\n") l) ++ "\\end{"++latexListType o++"}"++toLatex xs

latexFontOp Bold = "bf"
latexFontOp Emph = "em"
latexFontOp Mono = "tt"
latexFontOp Underline = "u"
latexFontOp Strike = "strike"

latexListType True = "enumeration"
latexListType False = "itemize"

output s f = case parse wikiParser "" s of
					Right n -> putStr (f n)
					Left e -> print e

htmlOutput s = output s toHtml
latexOutput s = output s toLatex
